// ListPFRO.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

namespace LeoHelpers
{
	// Returns false on failure. Use GetLastError().
	// hKeyParent -- hKey of parent under which szKeyPath should be created. Usually a hive like HKEY_CURRENT_USER
	// szKeyPath -- Sub key (can be a path) under hKeyHiveOrParent from which szValueName is read. Can be "" or NULL.
	// szValueName -- Name of value to query. Can be "" or NULL to read the unnamed default value.
	// sam -- Pass zero normally, or KEY_WOW64_32KEY or KEY_WOW64_64KEY if you need to access the alternative area. (KEY_QUERY_VALUE is specified automatically.)
	// pdwType -- Pointer to DWORD to receive type of data. NULL if unwanted.
	// plpData -- POINTER TO POINTER to receive data buffer. NULL if data unwanted.
	// pcbData -- Pointer to DWORD to receive size of data buffer. NULL if data unwanted.
	// You do NOT allocate a buffer, this function does it for you (like the stupid Registry API should have done, at least as an option).
	// If the function succeeds and plpData was not NULL you must delete[] *plpData when finished with it.
	// Like RegQueryValueEx, you may supply NULL plpData and non-NULL pcbData to get the required
	// buffer size (except for HKEY_PERFORMANCE_DATA), but again the function always allocates a buffer for you.
	// Although untested, reading HKEY_PERFORMANCE_DATA should work.
	// Pads whatever data comes out of the registry with 8 null bytes to avoid problems with badly stored data.
	bool RegQueryValue(HKEY hKeyParent, const wchar_t *szKeyPath, const wchar_t *szValueName, REGSAM sam, DWORD *pdwType, void **plpData, DWORD *pcbData);
};

bool LeoHelpers::RegQueryValue(HKEY hKeyParent, const wchar_t *szKeyPath, const wchar_t *szValueName, REGSAM sam,
									   DWORD *pdwType, void **plpData, DWORD *pcbData)
{
	// Documentation doesn't state that registry functions call SetLastError().
	// We need to preserve the error values they return through clean-up anyway.
	LONG regRes = ERROR_SUCCESS;

	HKEY hKeyToReadValueFrom;

	DWORD dwPcdDataStandin; // Used for getting the registry size if the user doesn't care about it.

	if (NULL == pcbData)
	{
		pcbData = &dwPcdDataStandin;
	}

	// szKeyPath may be NULL in which case another handle like hKeyParent is openned (and must be closed).
	regRes = ::RegOpenKeyEx(hKeyParent, szKeyPath, 0, sam | KEY_QUERY_VALUE, &hKeyToReadValueFrom);

	if (ERROR_SUCCESS == regRes)
	{
		if (NULL == plpData)
		{
			// They just want the type and/or the data size.
			regRes = ::RegQueryValueEx(hKeyToReadValueFrom, szValueName, 0, pdwType, NULL, pcbData);
		}
		else
		{
			// We add 8 bytes of zero padding to the end of any buffer we return. This is because
			// the registry does not check inputs to ensure they are the correct length for the
			// datatype and also does not ensure that strings are null-terminated and multi-strings are
			// double-null-terminated. Rather than add checks into all functions which call this one,
			// we pad with as many zero bytes as are needed to correct anything we might read.
			// 8 means a QUADDWORD that was zero length in the registry would still be safe.
			// 8 is also more than enough for double-null-terminator of a Unicode multi-string.
			const DWORD dwZeroPadding = 8;

			BYTE *allocatedBuffer = NULL;
			DWORD dwAllocatedBufferSize = 0; // For HKEY_PERFORMANCE_DATA, really.

			// Get the required buffer size by giving NULL for data buffer.
			regRes = ::RegQueryValueEx(hKeyToReadValueFrom, szValueName, 0, pdwType, NULL, pcbData);

			if (ERROR_SUCCESS == regRes)
			{
				dwAllocatedBufferSize = *pcbData;
				allocatedBuffer = new BYTE[dwAllocatedBufferSize + dwZeroPadding];
			}
			else if (ERROR_MORE_DATA == regRes)
			{
				// Most likely tried to read HKEY_PERFORMANCE_DATA. *pcdData is undefined at this point.
				*pcbData = 1024; // Must be a number x such that (Int(x/2) > 0)
				dwAllocatedBufferSize = *pcbData;
				allocatedBuffer = new BYTE[dwAllocatedBufferSize + dwZeroPadding];
			}
			else
			{
				*pcbData = 0;
				// leave allocatedBuffer NULL so the while loop does not execute.
			}

			while (NULL != allocatedBuffer)
			{
				regRes = ::RegQueryValueEx(hKeyToReadValueFrom, szValueName, 0, pdwType, allocatedBuffer, pcbData);

				if (ERROR_SUCCESS == regRes)
				{
					// Zero out the padding we added.
					ZeroMemory(allocatedBuffer + dwAllocatedBufferSize, dwZeroPadding);

					*plpData = allocatedBuffer;
					allocatedBuffer = NULL;
					break;
				}
				else if (ERROR_MORE_DATA == regRes)
				{
					// Try a larger buffer.
					delete [] allocatedBuffer;
					*pcbData = dwAllocatedBufferSize + (dwAllocatedBufferSize/2);
					dwAllocatedBufferSize = *pcbData;
					allocatedBuffer = new BYTE[dwAllocatedBufferSize + dwZeroPadding];
				}
				else
				{
					*pcbData = 0;
					delete [] allocatedBuffer;
					allocatedBuffer = NULL;
					break;
				}
			}
		}

		LONG closeRes = ::RegCloseKey(hKeyToReadValueFrom);
		if (ERROR_SUCCESS == regRes) { regRes = closeRes; }
	}

	if (ERROR_SUCCESS != regRes)
	{
		::SetLastError(regRes);
	}

	return(ERROR_SUCCESS == regRes);
}

int wmain(int argc, wchar_t* argv[])
{
	DWORD dwType = 0;
	WCHAR *szData = NULL;

	bool bTerse = false;

	if (argc == 2 && 0 == _wcsicmp(L"terse", argv[1]))
	{
		bTerse = true;
	}

	// The main value is PendingFileRenameOperations.
	//
	// PendingFileRenameOperations2 can also exist:
	//   "When there are a large number of files, the primary PendingFileRenameOperations REG_MULTI_SZ value becomes
	//   full and the system creates a PendingFileRenameOperations2 value which contains additional pending rename information."
	//   -- http://support.microsoft.com/kb/2014837
	//
	// PendingFileRenameOperations2 can exist either in addition to PendingFileRenameOperations or on its own.
	//
	// PendingFileRenameOperations3 etc. are not used by the OS (tested on Windows 8), so if the first two values fill
	// up then I guess you can't add any more pending operations.

	bool bGotAtLeastOne = false;
	bool bReportedError = false;

	for(int iKeyNum = 0; iKeyNum < 2; ++iKeyNum)
	{
		const wchar_t *szKeyName = (iKeyNum == 0) ? L"PendingFileRenameOperations" : L"PendingFileRenameOperations2";

		delete [] szData;
		szData = NULL;

		if (!LeoHelpers::RegQueryValue(HKEY_LOCAL_MACHINE, L"SYSTEM\\CurrentControlSet\\Control\\Session Manager", szKeyName, 0, &dwType, reinterpret_cast<void**>(&szData), NULL))
		{
			DWORD dwErr = GetLastError();

			if (ERROR_FILE_NOT_FOUND == dwErr)
			{
				continue; // Not found, skip it.
			}

			wprintf(L"Error %lu reading %s\n", dwErr, szKeyName);
			bReportedError = true;
			break;
		}

		if (REG_MULTI_SZ != dwType)
		{
			wprintf(L"%s not REG_MULTI_SZ\n", szKeyName);
			bReportedError = true;
			break;
		}

		if (L'\0' == *szData)
		{
			continue; // Empty, skip it.
		}

		bGotAtLeastOne = true;

		if (bTerse)
		{
			wprintf(L"Reboot\n");
			break; // Don't need to look for more in terse mode.
		}

		WCHAR *szNext = szData;

		while (L'\0' != (*szNext))
		{
			WCHAR *szSource = szNext;
			szNext += (wcslen(szNext) + 1);

			WCHAR *szDest = szNext;
			szNext += (wcslen(szNext) + 1);

			if (L'\0' == (*szDest))
			{
				wprintf(L"DELETE \"%s\"\n", szSource);
			}
			else
			{
				wprintf(L"MOVE \"%s\"\n to  \"%s\"\n", szSource, szDest);
			}
		}
	}

	delete [] szData;
	szData = NULL;

	if (!bGotAtLeastOne && !bTerse && !bReportedError)
	{
		wprintf(L"PFRO not found or empty.\n");
	}

	return 0;
}
